import datetime
import sys

MAX_DBLATENCY = 1 * (10**6) # in microseconds
SECONDS_IN_DAY = 86400

def canEvict(update):
    delta = datetime.datetime.now() - update.ts
    deltaInSeconds = (delta.days * SECONDS_IN_DAY) + delta.seconds
    deltaInMicroSeconds = deltaInSeconds*(10**6) + delta.microseconds
    return deltaInMicroSeconds > MAX_DBLATENCY

# update object used in Decat
class Update:
    def __init__(self, id, attrs, ts, requestId):
        # attrs in the form of key value pairs (dict)
        self.id = id
        self.attrs = attrs
        self.ts = ts
        self.requestId = requestId # request that caused this update
    
    def merge(self, other):
        if self.id == other.id:
            self.attrs.update(other.attrs)
            self.ts = max(self.ts, other.ts)

    def empty(self):
        return len(self.attrs) == 0

    def __str__(self):
        return str(self.id) + ": " + str(self.attrs) + "\n" + str(self.ts)


# used for LRUCache below
class Node:
    def __init__(self, value, prev=None, next=None):
        self.value = value
        self.prev = prev
        self.next = next

# LRUCache as used in Decat
# The update object might need to be modifief for MVCC-TSO
class LRUCache:
    dummyValue = -1

    def __init__(self, maxSize, canEvictFunc=canEvict):
        self.maxSize = maxSize
        self.size = 0
        self.dict = {}
        self.canEvictFunc = canEvictFunc

        # head and tail always point to dummy nodes
        self.head = Node(self.dummyValue)
        self.tail = Node(self.dummyValue)
        self.head.next = self.tail
        self.tail.prev = self.head

    def full(self):
        return self.size  == self.maxSize


    def debug(self):
        node = self.head.next
        while node != self.tail:
            print(node.value)
            print() 
            node = node.next

    # returns Update object defined at the top of this file
    def get(self, key):
        if key not in self.dict:
            return None
        else:
            return self.dict[key].value

    # key is obj id
    # value is an Update object defined at the top of this file
    def put(self, update):
        key = update.id

        if key in self.dict:
            # id present in Cache; just update it
            node = self.dict[key]
            node.value.merge(update)

            # remove it from where it is
            node.prev.next = node.next
            node.next.prev = node.prev
            
            # reattach it at head
            node.next = self.head.next
            node.prev = self.head

            self.head.next.prev = node
            self.head.next = node

            return node.value
        else:
            # id not present
            if self.full():
                if not self.evict():
                    print("Unable to evict entry from LRUCache: try increasing maxSize?")
                    return None
            
            # insert in list
            node = Node(update, self.head, self.head.next)
            self.head.next.prev = node
            self.head.next = node
            
            self.size += 1
            self.dict[key] = node

            return node.value

    
    # don't use this; this removes the whole update object (all attrs)
    # this might not be the behaviour you want
    # key is actually the id of object (subject/resource)
    def remove(self, key):
        if key not in self.dict:
            return False
        else:
            node = self.dict[key]
            node.prev.next = node.next
            node.next.prev = node.prev
            self.size -= 1
            del self.dict[key]
            return True
        

    # starting from tail 
    # go through the list until canEvictFunc says True
    def evict(self):
        node = self.tail.prev
        while node != self.head:
            if self.canEvictFunc(node.value):
                node.prev.next = node.next
                node.next.prev = node.prev
                self.size -= 1
                return True
            node = node.prev
        return False

# used this for tentative updates
# in Decat
class Cache:
    def __init__(self):
        self.cache = dict()

    def put(self, update):
        requestId = update.requestId
        key = update.id
        if key not in self.cache:
            self.cache[key] = {}
        
        self.cache[key][requestId] = update
        return self.cache[key][requestId]

    # returns a dict {<requestId>: updateObject1, ...}
    def get(self, id):
        if id in self.cache:
            return self.cache[id]
        else:
            return None

    def remove(self, id, requestId):
        if id in self.cache and requestId in self.cache[id]:
            del self.cache[id][requestId]
            return True
        else:
            return False
    
def debug1(d):
    for k, v in d.items():
        print(k, v)